@@ -28,6 +28,7 @@ import socket
 import httplib
 import urllib
 import https
+import time
 
 import config
 import utils
@@ -35,14 +36,16 @@ import remote
 import local
 
 
-VERSION = '0.3.0'
+VERSION = '0.3.2'
 
 
 def hook_setup(cmd):
+    if '-h' in cmd: return
     config.setup()
     sys.exit(0)
 
 def hook_destroy(cmd):
+    if '-h' in cmd: return
     local.confirm('Please confirm destruction')
 
 def hook___version(cmd):
@@ -68,6 +71,12 @@ def hook___import(cmd):
         _import(sys.stdin)
     sys.exit(0)
 
+_trace = False
+def hook___trace(cmd):
+    global _trace
+    cmd.pop(0)
+    _trace = True
+
 def run_remote(cmd):
     r = remote.Remote()
     handlers = {
@@ -81,6 +90,7 @@ def run_remote(cmd):
             'git': r.git,
             'hg': r.hg,
             'upload': r.upload,
+            'loop': lambda *x: run_loop(*x),
             'confirm': local.confirm,
             'call': lambda x: run_command(x, True),
             'echo': lambda x: sys.stdout.write('{0}\n'.format(x)),
@@ -115,8 +125,26 @@ def sign_request(api_key, method, request):
             }
     return headers
 
+def run_loop(cmd, state, interval, timeout=100):
+    count = 0
+    while True:
+        cmd[-1] = state
+        more, state = run_command(cmd, True, retval=True, progress=True)
+        if more is False:
+            break
+        if count > timeout:
+            sys.stderr.write('\nTimed out\n')
+            sys.exit(1)
+        if isinstance(state, dict) and 'active' in state:
+            count = 0
+        else:
+            count += 1
+        if interval > 0:
+            time.sleep(interval)
+    return
+
 _request = None
-def run_command(cmd, internal=False):
+def run_command(cmd, internal=False, retval=False, progress=False):
     global _request
     if internal is False and not os.path.exists(config.CONFIG_KEY):
         run_command('get_key', True)
@@ -131,38 +159,61 @@ def run_command(cmd, internal=False):
         else:
             req = httplib.HTTPConnection(url['host'], int(url['port'] or 80))
         _request = req
+    original_cmd = cmd
     cmd = urllib.quote(json.dumps(cmd))
-    query = '{0}run?{1}={2}'.format(url['path'], 'c' if internal is True else 'q', cmd)
+    query = '{0}run?{1}={2}'.format(url['path'] or '/', 'c' if internal is True else 'q', cmd)
     headers = sign_request(cfg.apikey, 'GET', query)
     headers.update({
         'User-Agent': 'dotcloud/cli (version: {0})'.format(VERSION),
         'X-DotCloud-Version': VERSION
         })
+    trace_id = None
     try:
         req.request('GET', query, headers=headers)
         resp = req.getresponse()
         info = resp.getheader('X-Dotcloud-Info')
+        trace_id = resp.getheader('X-Dotcloud-TraceID')
         data = resp.read()
+        req.close()
         if _export is True:
             print data
             return
         if info:
             utils.warning(info.replace(';', '\n'))
+        if _trace and trace_id:
+            utils.info('TraceID for "{0}": {1}'.format(
+                original_cmd, trace_id))
         data = json.loads(data)
         if data['type'] == 'cmd':
             return run_remote(data['data'])
-        if 'data' in data:
+        if 'data' in data and len(data['data']) > 0:
+            if progress:
+                sys.stderr.write('\r')
             print data['data']
+        elif progress:
+            sys.stderr.write('.')
     except socket.error, e:
-        utils.die('Cannot reach DotCloud service ("{0}").\nPlease check the connectivity and try again.'.format(str(e)))
+        utils.die('Cannot reach DotCloud service ("{0}").\n' \
+                'Please check the connectivity and try again.'.format(str(e)))
     except Exception, e:
-        utils.die('DotCloud service unavailable ("{0}").\nPlease try again later. If the problem persists, send an email to support@dotcloud.com.'.format(e))
+        message = 'DotCloud service unavailable ("{0}").\n' \
+                'Data received:\n---\n{1}\n---\n' \
+                'Please try again later. ' \
+                'If the problem persists, send an email to support@dotcloud.com'.format(
+                    e, data)
+        if trace_id:
+            message += '\nand include the following TraceID: {0}'.format(trace_id)
+        else:
+            message += '.'
+        utils.die(message)
     finally:
         try:
             req.close()
         except Exception:
             pass
     if internal is True:
+        if retval is True and 'retval' in data:
+            return data['retval']
         return
     # Return proper exit code to the shell
     if not data:
