@@ -19,37 +19,32 @@
 ## THE SOFTWARE.
 
 import os
-import time
 import subprocess
 
 import utils
 import config
-
+import local
 
 class Remote(object):
 
     def __init__(self):
         self._verbose = True
         self._ssh_master = None
-
-    def __del__(self):
-        if self._ssh_master is not None and self._ssh_master.poll() is None:
-            self._ssh_master.terminate()
-
-    def set_url(self, url):
-        parts = utils.parse_url(url)
-        (self._user, self._host, self._port) = (parts['user'] or 'dotcloud', parts['host'], parts['port'] or '22')
-        self._url = url
-        self._ssh_sock = '/tmp/dotcloud-{0}.sock'.format(time.time())
         self._ssh_options = (
-                'ssh',
+                'ssh', '-t',
                 '-i', config.CONFIG_KEY,
+                '-o', 'LogLevel=QUIET',
                 '-o', 'UserKnownHostsFile=/dev/null',
                 '-o', 'StrictHostKeyChecking=no',
                 '-o', 'PasswordAuthentication=no',
                 '-o', 'ServerAliveInterval=10'
                 )
 
+    def set_url(self, url):
+        parts = utils.parse_url(url)
+        (self._user, self._host, self._port) = (parts['user'] or 'dotcloud', parts['host'], parts['port'] or '22')
+        self._url = url
+
     def set_verbose(self, flag):
         self._verbose = flag
 
@@ -66,54 +61,25 @@ class Remote(object):
         utils.warning('Please try again. If the problem persists, send an email to support@dotcloud.com.')
         utils.warning('Also please check that your are allowed to make an SSH connection to a custom port.')
 
-    def _spawn_ssh_master(self):
-        if self._ssh_master and self._ssh_master.poll() is None:
-            return
-        try:
-            self._ssh_master = subprocess.Popen(
-                    self._ssh_options + (
-                        '-p', self._port,
-                        '-l', self._user,
-                        '-S', self._ssh_sock,
-                        '-M',
-                        self._host, 'echo ok ; cat'
-                        ),
-                    stdin=subprocess.PIPE,
-                    stdout=subprocess.PIPE,
-                    stderr=subprocess.PIPE
-            )
-            if self._ssh_master.stdout.read(3) != 'ok\n':
-                ret = self._ssh_master.poll()
-                if ret is not None and ret != 0:
-                    self.warning_ssh()
-                utils.die('SSH connection failed on "{0}"'.format(self._url))
-        except OSError:
-            self.die('ssh')
-
     def _escape(self, s):
         for c in ('`', '$', '"'):
             s = s.replace(c, '\\' + c)
         return s
 
-    def _ssh(self, cmd, tty=True, **kwargs):
-        self._spawn_ssh_master()
+    def _ssh(self, cmd, **kwargs):
         p_args = self._ssh_options + (
-                '-t' if tty else '-T',
                 '-l', self._user,
                 '-p', self._port,
-                '-S', self._ssh_sock,
                 self._host,
                 'bash -l -c "{0}"'.format(self._escape(cmd))
                 )
         return subprocess.Popen(p_args, **kwargs)
 
     def _scp(self, src, dest):
-        self._spawn_ssh_master()
         scp = (
                 'scp', '-P', self._port, '-r',
                 '-o', 'UserKnownHostsFile=/dev/null',
                 '-o', 'StrictHostKeyChecking=no',
-                '-o', 'ControlPath={0}'.format(self._ssh_sock),
                 src, dest
         )
         return subprocess.call(scp, close_fds=True)
@@ -121,16 +87,15 @@ class Remote(object):
     def key(self, data):
         with open(config.CONFIG_KEY, 'w') as f:
             f.write(data)
-            os.fchmod(f.fileno(), 0600)
+            if not utils.is_windows():
+                os.fchmod(f.fileno(), 0600)
 
     def sftp(self):
-        self._spawn_ssh_master()
         sftp = (
                 'sftp',
                 '-o', 'Port={0}'.format(self._port),
                 '-o', 'UserKnownHostsFile=/dev/null',
                 '-o', 'StrictHostKeyChecking=no',
-                '-o', 'ControlPath={0}'.format(self._ssh_sock),
                 '{user}@{host}'.format(user=self._user, host=self._host)
                 )
         return subprocess.call(sftp, close_fds=True)
@@ -149,7 +114,7 @@ class Remote(object):
         return self._ssh(cmd).wait()
 
     def run_script(self, script):
-        proc = self._ssh('/bin/bash', tty=False, stdin=subprocess.PIPE)
+        proc = self._ssh('/bin/bash', stdin=subprocess.PIPE)
         proc.stdin.write(script)
         proc.communicate()
 
@@ -159,15 +124,17 @@ class Remote(object):
         url = utils.parse_url(destination)
         ssh = ' '.join(self._ssh_options)
         ssh += ' -p {0}'.format(url['port'])
-        if not local_dir.endswith('/'):
+        if not os.path.isfile(local_dir) and not local_dir.endswith('/'):
             local_dir += '/'
-        rsync = (
-                    'rsync', '-lpthrvz', '--delete', '--safe-links',
-                ) + tuple('--exclude={0}'.format(e) for e in excludes) + (
-                    '-e', ssh, local_dir,
-                    '{user}@{host}:{dest}/'.format(user=url['user'],
-                        host=url['host'], dest=url['path'])
-                )
+        ignore_file = os.path.join(local_dir, '.dotcloudignore')
+        ignore_opt = ('--exclude-from', ignore_file)
+        if not os.path.exists(ignore_file):
+            ignore_opt = tuple()
+        rsync = ('rsync', '-lpthrvz', '--delete', '--safe-links') + \
+                tuple('--exclude={0}'.format(e) for e in excludes) + ignore_opt + \
+                ('-e', ssh, local_dir,
+                        '{user}@{host}:{dest}/'.format(user=url['user'],
+                        host=url['host'], dest=url['path']))
         try:
             ret = subprocess.call(rsync, close_fds=True)
             if ret != 0:
@@ -194,6 +161,7 @@ class Remote(object):
         with utils.cd(local_dir):
             try:
                 os.environ['GIT_SSH'] = '__dotcloud_git_ssh'
+                os.environ['DOTCLOUD_SSH_KEY'] = config.CONFIG_KEY
                 ret = subprocess.call(('git', 'push', '-f', '--all',
                     destination), close_fds=True)
                 if ret != 0:
@@ -202,10 +170,49 @@ class Remote(object):
             except OSError:
                 self.die('git')
 
-    def upload(self, local_dir, destination, args):
-        self.info('# upload {0} {1}'.format(local_dir, destination))
+    def upload_method(self, local_dir):
         if os.path.isdir(os.path.join(local_dir, '.hg')):
-            return self.hg(local_dir, destination, args.get('hg', {}))
+            return 'hg'
         if os.path.isdir(os.path.join(local_dir, '.git')):
+            return 'git'
+        return 'rsync'
+
+    def upload(self, local_dir, destination, args):
+        if args.get('check'):
+            local_dir = self.check_pushdir(local_dir)
+        if args.get('verify_key'):
+            self.verify_key(args.get('verify_key'))
+        self.info('# upload {0} {1}'.format(local_dir, destination))
+        method = args.get('force_method') or self.upload_method(local_dir)
+        if method == 'hg':
+            return self.hg(local_dir, destination, args.get('hg', {}))
+        if method == 'git':
             return self.git(local_dir, destination, args.get('git', {}))
         return self.rsync(local_dir, destination, args.get('rsync', {}))
+
+    def verify_key(self, key):
+        f = open(config.CONFIG_KEY).read()
+        if not f.replace('\n', '') == key.replace('\n', ''):
+            utils.die('Error: you seem to have a stale key file "{0}"\n'
+                      'Remove the file or run "dotcloud setup" command again.'.format(config.CONFIG_KEY))
+
+    def check_pushdir(self, local_dir):
+        orig = dir = os.path.realpath(local_dir)
+        if ':' in orig: # win32?
+            return self.check_pushdir_win32(local_dir)
+        while True:
+            if os.path.isdir(dir) and os.path.exists(os.path.join(dir, 'dotcloud.yml')):
+                if dir != orig:
+                    utils.info('# Found dotcloud.yml: Using {0} as a base directory'.format(dir))
+                return dir
+            prev = dir
+            dir = os.path.realpath(os.path.join(dir, os.path.pardir))
+            if dir == prev:
+                local.confirm('Could not find dotcloud.yml file in {0} and parent directories. Proceed?'.format(orig))
+                break
+        return orig
+
+    def check_pushdir_win32(self, local_dir):
+        if not os.path.exists(os.path.join(local_dir, 'dotcloud.yml')):
+            local.confirm('Could not find dotcloud.yml file in {0}. Proceed?'.format(local_dir))
+        return local_dir
